/*

    File: fat.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef DJGPP
#include <curses.h>
#else
#include <ncurses.h>
#endif
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "interface.h"
#define MAX_TBFAT 10
int set_FAT_info(t_param_disk_cst disk_car,struct fat_boot_sector *fat_header, t_diskext partition,const int debug);
int set_FAT32_part_name(t_param_disk_cst disk_car, t_diskext partition, struct fat_boot_sector*fat_header);
uint sector_size(const struct fat_boot_sector *fat_header);
uint get_dir_entries(const struct fat_boot_sector *fat_header);
uint sectors(const struct fat_boot_sector *fat_header);
t_fat_test start_fat[MAX_TBFAT];

int dump_fat_rapport(const struct fat_boot_sector*fh1, int part_type)
{
  ecrit_rapport("sector_size  %u\n", sector_size(fh1));
  ecrit_rapport("cluster_size %u\n", fh1->cluster_size);
  ecrit_rapport("reserved     %u\n", fh1->reserved);
  ecrit_rapport("fats         %u\n", fh1->fats);
  ecrit_rapport("dir_entries  %u\n", get_dir_entries(fh1));
  ecrit_rapport("sectors      %u\n", sectors(fh1));
  ecrit_rapport("media        %02X\n", fh1->media);
  ecrit_rapport("fat_length   %u\n", fh1->fat_length);
  ecrit_rapport("secs_track   %u\n", fh1->secs_track);
  ecrit_rapport("heads        %u\n", fh1->heads);
  ecrit_rapport("hidden       %u\n", fh1->hidden);
  ecrit_rapport("total_sect   %u\n", fh1->total_sect);
  if(part_type==P_32FAT_1)
  {
	ecrit_rapport("fat32_length %u\n", fh1->fat32_length);
	ecrit_rapport("flags        %04X\n", fh1->flags);
	ecrit_rapport("version      %u.%u\n", fh1->version[0], fh1->version[1]);
	ecrit_rapport("root_cluster %u\n", fh1->root_cluster);
	ecrit_rapport("info_sector  %u\n", fh1->info_sector);
	ecrit_rapport("backup_boot  %u\n", fh1->backup_boot);
  }
  return 0;
}

int dump_2fat_rapport(const struct fat_boot_sector*fh1, const struct fat_boot_sector*fh2, int part_type)
{
  ecrit_rapport("\nsector_size  %u %u", sector_size(fh1),sector_size(fh2));
  ecrit_rapport("\ncluster_size %u %u", fh1->cluster_size,fh2->cluster_size);
  ecrit_rapport("\nreserved     %u %u", fh1->reserved,fh2->reserved);
  ecrit_rapport("\nfats         %u %u", fh1->fats,fh2->fats);
  ecrit_rapport("\ndir_entries  %u %u", get_dir_entries(fh1),get_dir_entries(fh2));
  ecrit_rapport("\nsectors      %u %u", sectors(fh1),sectors(fh2));
  ecrit_rapport("\nmedia        %02X %02X", fh1->media,fh2->media);
  ecrit_rapport("\nfat_length   %u %u", fh1->fat_length,fh2->fat_length);
  ecrit_rapport("\nsecs_track   %u %u", fh1->secs_track,fh2->secs_track);
  ecrit_rapport("\nheads        %u %u", fh1->heads,fh2->heads);
  ecrit_rapport("\nhidden       %u %u", fh1->hidden,fh2->hidden);
  ecrit_rapport("\ntotal_sect   %u %u", fh1->total_sect,fh2->total_sect);
  if(part_type==P_32FAT_1)
  {
	ecrit_rapport("\nfat32_length %u %u", fh1->fat32_length,fh2->fat32_length);
	ecrit_rapport("\nflags        %04X %04X", fh1->flags,fh2->flags);
	ecrit_rapport("\nversion      %u.%u  %u.%u", fh1->version[0], fh1->version[1],fh2->version[0], fh2->version[1]);
	ecrit_rapport("\nroot_cluster %u %u", fh1->root_cluster,fh2->root_cluster);
	ecrit_rapport("\ninfo_sector  %u %u", fh1->info_sector,fh2->info_sector);
	ecrit_rapport("\nbackup_boot  %u %u", fh1->backup_boot,fh2->backup_boot);
  }
  ecrit_rapport("\n");
  return 0;
}

int check_FAT(t_param_disk_cst disk_car,t_diskext partition,const int debug)
{
  t_sector buffer;
//  recover_FAT_CIH(disk_car,partition,debug,0);
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba))
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  if(test_FAT(disk_car,(struct fat_boot_sector *)buffer,partition,debug,0))
  {
    if(debug)
    {
      ecrit_rapport("\n\ntest_FAT()\n");
      aff_part_rapport(2,disk_car,partition);
      dump_fat_rapport((const struct fat_boot_sector*)buffer, partition->part_type);
    }
    return 1;
  }
  set_FAT_info(disk_car,(struct fat_boot_sector *)buffer,partition,debug);
/*  aff_buffer(BUFFER_ADD,"Ok\n"); */
  return 0;
}

int set_FAT_info(t_param_disk_cst disk_car,struct fat_boot_sector *fat_header, t_diskext partition,const int debug)
{
  switch(partition->part_type)
  {
    case P_12FAT:
    case P_12FATH:
    case P_16FAT:
    case P_16FATH:
    case P_16FATBD:
    case P_16FATBDH:
      set_part_name(partition,((const char*)fat_header)+FAT1X_PART_NAME,11);
      break;
    case P_32FAT_1:
    case P_32FAT_2:
    case P_32FAT_1H:
    case P_32FAT_2H:
      set_FAT32_part_name(disk_car,partition,fat_header);
      break;
  }
  return 0;
}

/* The idea is to be able to rebuild boot sector and first FAT
 * when they have been erased by CIH virus */
unsigned int get_frag(uchar *buffer,int fat_type)
{
  int nb_frag=1;
  switch(fat_type)
  {
    case 12:
      nb_frag=0x200;
    case 16:
      {
	int i;
	word *p16=(word*)buffer;
	for(i=1; i<0x200/2-1; i++)
	  if(p16[i]!=0xFFFF && p16[i]!=i+1)
	    nb_frag++;
      }
      break;
    case 32:
      {
	unsigned int i;
	dword *p32=(dword*)buffer;
      	for(i=1; i<0x200/4-1; i++)
      	  if(p32[i]!=(unsigned)0xFFFFFF0F && p32[i]!=i+1)
	    nb_frag++;
      }
      break;
  }
  return nb_frag;
}


int check_entree(uchar *entree)
{ /* 1 si non nulle */
  int i;
  for(i=0;i<0x20;i++)
	if(*(entree+i))
	  return 1;
  return 0;
}

dword get_subdirectory(t_param_disk disk_car,const loff_t hd_offset,const dword i)
{
  uchar buffer[0x200];
  if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
  {
	ecrit_rapport(m_ROOT_CLUSTER_RERR); return 1;
  }
  /*                     12345678123*/
  if(strncmp(&buffer[0],".          ",8+3)!=0)
	return 1;
  if((unsigned)((buffer[0x15]<<24)|(buffer[0x14]<<16)|(buffer[0x1B]<<8)|buffer[0x1A])!=i)
	return 1;
  /*                        12345678123*/
  if(strncmp(&buffer[0x20],"..         ",8+3)!=0)
	return 1;
  return (buffer[0x35]<<24)+(buffer[0x34]<<16)+(buffer[0x3B]<<8)+buffer[0x3A];
}

dword get_next_cluster(t_param_disk disk_car,t_diskext_cst partition, int part_type,const int offset, const dword cluster)
{
  /* Offset can be offset to FAT1 or to FAT2 */
  uchar buffer[0x200];
  switch(part_type)
  {
	case P_16FAT:
	  {
		/* secteur 1: cluster X 2-256 */
		/* secteur 2: cluster 257-513 */
		word *p16=(word*)&buffer;
		dword offset_s,offset_o;
		offset_s=(cluster-1)/256;
		offset_o=cluster-1-offset_s*256;
		if(hd_read_LBA(disk_car,1, &buffer, partition->lba+offset+offset_s))
		{
		  ecrit_rapport("\nget_next_cluster error"); return 0;
		}
//		ecrit_rapport("cluster=%u offset_s=%u offset_o=%u",cluster,offset_s,offset_o);
		return p16[offset_o];
	  }
	case P_32FAT_1:
	  {
		dword *p32=(dword*)&buffer;
		dword offset_s,offset_o;
		offset_s=(cluster-1)/128;
		offset_o=cluster-1-offset_s*128;
		if(hd_read_LBA(disk_car,1, &buffer, partition->lba+offset+offset_s))
		{
		  ecrit_rapport("\nget_next_cluster error"); return 0;
		}
//		ecrit_rapport("cluster=%u offset_s=%u offset_o=%u",cluster,offset_s,offset_o);
		return p32[offset_o];
	  }
	  break;
  }
  return 0;
}

dword get_prev_cluster(t_param_disk disk_car,t_diskext_cst partition, int part_type,const int offset, const dword cluster, const dword no_of_cluster)
{
  dword prev_cluster;
  for(prev_cluster=2;prev_cluster<=no_of_cluster;prev_cluster++)
  {
	if(get_next_cluster(disk_car,partition,part_type,offset, prev_cluster)==cluster)
	  return prev_cluster;
  }
  return 0;
}

int get_root_cluster(t_param_disk disk_car,t_diskext_cst partition,const int cluster_size, const dword no_of_cluster,const int offset)
{
  dword root_cluster;
  uchar buffer[cluster_size*0x200];
  for(root_cluster=2;root_cluster<no_of_cluster;root_cluster++)
  {
	if(hd_read_LBA(disk_car,cluster_size, &buffer, partition->lba+offset+(root_cluster-2)*cluster_size))
	{
	  ecrit_rapport(m_ROOT_CLUSTER_RERR); return 0;
	}
	if(strncmp(&buffer[0],".          ",8+3)!=0)
	{
	  int i,found=1;
	  int etat=0,nb_subdir=0;
	  for(i=0;found && (i<16*cluster_size);i++)
	  {
		if(check_entree(&buffer[i*0x20])==0)
		{
		  if(etat==0)
			etat=1;
		}
		else
		  if(etat==1)
		  {
			etat=2;
			found=0;
		  }
		if(buffer[i*0x20+0xB]==0x10) /* Test directory */
		{
		  dword cluster=(buffer[i*0x20+0x15]<<24)+(buffer[i*0x20+0x14]<<16)+
			(buffer[i*0x20+0x1B]<<8)+buffer[i*0x20+0x1A];
		  nb_subdir++;
		  if((cluster>no_of_cluster)||(get_subdirectory(disk_car,partition->lba+offset+(cluster-2)*cluster_size,cluster)!=0))
			found=0;
		}
	  }
	  if(nb_subdir<3)
		found=0;
//	  ecrit_rapport("\ncluster %d, etat=%d, found=%d,nb_subdir=%d",root_cluster,etat,found,nb_subdir);
	  if(found)
	  {
		dword tmp;
		/* Il faut ajouter un parcours arriere de la FAT 
		 * car on localise le dernier cluster du root_cluster
		 * */
		tmp=get_prev_cluster(disk_car,partition,partition->part_type,offset,root_cluster,no_of_cluster);
		while(tmp)
		{
		  root_cluster=tmp;
		tmp=get_prev_cluster(disk_car,partition,partition->part_type,offset,root_cluster,no_of_cluster);
		}
		return root_cluster;
	  }
	}
  }
  return 0;
}

int analyse_dir_entries(t_param_disk disk_car,t_diskext_cst partition, const int offset)
{
  uchar buffer[0x200];
  int i,j;
  int etat=0;
  loff_t hd_offset;
  hd_offset=partition->lba+offset;
  for(i=0;(i<200)&&(etat!=2);i++)
  {
	if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
	{
	  ecrit_rapport(m_ANALYSE_DIR_FAT_RERR); return 0;
	}
	for(j=0;j<16;j++)
	{
	  if(check_entree(&buffer[j*0x20])==0)
	  {
		if(etat==0)
		{
		  ecrit_rapport("\ndir_entries 0->1 %d",i*16+j);
		  etat=1;
		}
	  }
	  else
		if(etat==1)
		{
		  ecrit_rapport("\ndir_entries 1->2 %d",i*16+j);
		  etat=2;
		}
	}
	hd_offset++;
  }
  return (i-1)*16;
}

int create_boot_sector(t_param_disk disk_car,t_diskext_cst partition, int reserved, int debug,int dir_entries,dword root_cluster,int cluster_size,int fat_length)
{
  uchar newboot[0x200];
  struct fat_boot_sector *fat_header=(struct fat_boot_sector *)&newboot;
  memset(&newboot,0,0x200);
  fat_header->sector_size[0]=0x200 & 0xFF;
  fat_header->sector_size[1]=0x200>>8;
  fat_header->fats=2;
  fat_header->media=0xF8;
  fat_header->secs_track=disk_car->CHS.sector;
  fat_header->heads=disk_car->CHS.head+1;
  fat_header->marker=0xAA55;
  fat_header->hidden=partition->lba;
  fat_header->cluster_size=cluster_size;
  fat_header->reserved=reserved;
  if(partition->part_size>0xFFFF)
  {
	fat_header->sectors[0]=0;
	fat_header->sectors[1]=0;
	fat_header->total_sect=partition->part_size;
  }
  else
  {
  	fat_header->sectors[0]=partition->part_size&0xFF;
	fat_header->sectors[1]=(partition->part_size>>8)&0xFF;
	fat_header->total_sect=0;
  }
  switch(partition->part_type)
  {
	case P_12FAT:
	case P_12FATH:
	  fat_header->fat_length=fat_length;
	  fat_header->dir_entries[0]=dir_entries & 0xFF;
	  fat_header->dir_entries[1]=(dir_entries >> 8) & 0xFF;
	  break;
	case P_16FAT:
	  fat_header->fat_length=fat_length;
	  fat_header->dir_entries[0]=dir_entries & 0xFF;
	  fat_header->dir_entries[1]=(dir_entries >> 8) & 0xFF;
	  break;
	case P_32FAT_1:
	  fat_header->fat32_length=fat_length;
	  fat_header->info_sector=1;
	  fat_header->backup_boot=6;
	  fat_header->root_cluster=root_cluster;
	  break;
	default:
	  ecrit_rapport("create_boot_sector severe error\n");
	  exit(1);
  }
  if(debug)
  {
	uchar orgboot[0x200];
	struct fat_boot_sector *org_fat_header=(struct fat_boot_sector *)&orgboot;
	if(hd_read_LBA(disk_car,1, &orgboot, partition->lba+fat_header->backup_boot))
	{
	  ecrit_rapport(m_CHKFAT_RERR); return 1;
	}
	ecrit_rapport("\nInformation from new boot sector");
	ecrit_rapport("\nDisk %02X",partition->disk);
	aff_part_rapport(1,disk_car,partition);
	dump_2fat_rapport(fat_header,org_fat_header,partition->part_type);
  }
  /* Need to write boot sector, copy FAT, and create FAT32 things */
//  hd_write_LBA(disk_car,1, &newboot, partition->lba);
//  if(partition->part_type==P_32FAT_1)
//  hd_write_LBA(disk_car,1, &newboot, partition->lba+fat_header->backup_boot);
  return 0;
}

int calcul_cluster_size(int part_type,dword data_size,int fat_length)
{
  switch(part_type)
  {
	case P_12FAT:
	  return up2power(data_size/((fat_length-1)*512/1.5-1));
	case P_16FAT:
	  return up2power(data_size/((fat_length-1)*512/2-1));
	case P_32FAT_1:
	  return up2power(data_size/((fat_length-1)*512/4-1));
  }
  return 0;
}

int search_fat_aux(t_param_disk disk_car,t_diskext_cst partition,dword max_offset,int p_fat12,int p_fat16,int p_fat32,int *f_fat12,int*f_fat16,int*f_fat32,int debug,int dump_ind)
{
  dword offset;
  int pos_fat_test=0;
  loff_t hd_offset;
  hd_offset=partition->lba;
  if(debug)
	ecrit_rapport("search_fat_aux(partition,max_offset=%d,p_fat12=%d,p_fat16=%d,p_fat32=%d,*f_fat12,*f_fat16,*f_fat32,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
  for(offset=0;(offset<max_offset)&&(pos_fat_test<MAX_TBFAT);offset++)
  {
    uchar buffer[0x200];
    int inc_ind=0;
    move(ANALYSE_Y,ANALYSE_X);
    clrtoeol();	/* before addstr for BSD compatibility */
    doprintf("search_fat_aux hd_offset %10d/%d",hd_offset,max_offset);
    if(hd_read_LBA(disk_car,1, &buffer, hd_offset))
    {
      continue;
    }
    if(buffer[0]==0xF8 && buffer[1]==0xFF)
    {
      start_fat[pos_fat_test].offset=offset;
      start_fat[pos_fat_test].f_fat12=0;
      start_fat[pos_fat_test].f_fat16=0;
      start_fat[pos_fat_test].f_fat32=0;
      if(p_fat32==1 && buffer[2]==0xFF && buffer[3]==0x0F && buffer[4]==0xFF && buffer[5]==0xFF && buffer[6]==0xFF && buffer[7]==0xFF)
      {
	start_fat[pos_fat_test].f_fat32=get_frag(buffer,32);
	if(*f_fat32==0 || start_fat[pos_fat_test].f_fat32<*f_fat32)
	  *f_fat32=start_fat[pos_fat_test].f_fat32;
	inc_ind=1;
      }
      if(p_fat16==1 && buffer[2]==0xFF)
      {
	start_fat[pos_fat_test].f_fat16=get_frag(buffer,16);
	if(*f_fat16==0 || start_fat[pos_fat_test].f_fat16<*f_fat16)
	  *f_fat16=start_fat[pos_fat_test].f_fat16;
	inc_ind=1;
      }
      if(p_fat12==1)
      {
	start_fat[pos_fat_test].f_fat12=get_frag(buffer,12);
	if(*f_fat12==0 || start_fat[pos_fat_test].f_fat12<*f_fat12)
	  *f_fat12=start_fat[pos_fat_test].f_fat12;
	inc_ind=1;
      }
      if(inc_ind)
      {
	if(dump_ind)
	  dump(buffer,0x200);
	if(debug)
	  ecrit_rapport("%5d  12:%d 16:%d 32:%d\n",start_fat[pos_fat_test].offset,start_fat[pos_fat_test].f_fat12,start_fat[pos_fat_test].f_fat16,start_fat[pos_fat_test].f_fat32);
	pos_fat_test++;
      }
    }
    hd_offset++;
  }
  move(ANALYSE_Y,ANALYSE_X);
  clrtoeol();
  return pos_fat_test;
}

int get_part_type(int f_fat12,int f_fat16,int f_fat32,int debug)
{
  int part_type=0;
  if(debug)
    ecrit_rapport("get_part_type(f_fat12=%d,f_fat16=%d,f_fat32=%d,debug=%d) => ",f_fat12,f_fat16,f_fat32,debug);
  /* Election du type de FAT */
  if(f_fat12)
  {
	if(f_fat16)
	{
	  if(f_fat12<f_fat16)
		part_type=P_12FAT;
	  else
		part_type=P_16FAT;
	}
	else
	  part_type=P_12FAT;
  }
  else
  {
	if(f_fat16)
	{
	  if(f_fat32)
	  {
		if(f_fat16<f_fat32)
		  part_type=P_16FAT;
		else
		  part_type=P_32FAT_1;
	  }
	  else
		part_type=P_16FAT;
	}
	else
	  if(f_fat32) part_type=P_32FAT_1;
  }
  if(debug)
  {
    switch(part_type)
    {
      case P_12FAT: ecrit_rapport("FAT12\n"); break;
      case P_16FAT: ecrit_rapport("FAT16\n"); break;
      case P_32FAT_1: ecrit_rapport("FAT32\n"); break;
      default: ecrit_rapport("No FAT\n"); break;
    }
  }
  return part_type;
}

int select_fat(int pos_fat_test, int part_type, int f_fat12, int f_fat16, int f_fat32, int debug)
{
  int i,pos=0;
  switch(part_type)
  {
	case P_12FAT:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat12==f_fat12)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat12=start_fat[i].f_fat12;
		  if(debug)
			ecrit_rapport("\nFAT12 at %u",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
	case P_16FAT:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat16==f_fat16)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat16=start_fat[i].f_fat16;
		  if(debug)
			ecrit_rapport("\nFAT16 at %u",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
	case P_32FAT_1:
	  for(i=0;i<pos_fat_test;i++)
	  {
		if(start_fat[i].f_fat32==f_fat32)
		{
		  start_fat[pos].offset=start_fat[i].offset;
		  start_fat[pos].f_fat32=start_fat[i].f_fat32;
		  if(debug)
			ecrit_rapport("\nFAT32 at %u",start_fat[pos].offset);
		  pos++;
		}
	  }
	  break;
  }
  return pos;
}

int set_fat_offset(t_param_disk disk_car,int*reserved, int*fat_length, t_diskext_cst partition,dword max_offset,int p_fat12,int p_fat16,int p_fat32,int debug,int dump_ind)
{
  int pos_fat_test=0;
  int f_fat12=0, f_fat16=0, f_fat32=0;
  int part_type;
  if(debug)
	ecrit_rapport("set_fat_offset(*reserved,*fat_length,partition,max_offset=%d,p_fat12=%d,p_fat16=%d,p_fat32=%d,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
  pos_fat_test=search_fat_aux(disk_car,partition,max_offset,p_fat12,p_fat16,p_fat32,&f_fat12,&f_fat16,&f_fat32,debug,dump_ind);
  part_type=get_part_type(1.5*f_fat12, 2*f_fat16, 4*f_fat32,debug);
  pos_fat_test=select_fat(pos_fat_test,part_type,f_fat12,f_fat16,f_fat32,debug);

  switch(part_type)
  {
	case P_12FAT:
	case P_16FAT:
	  *reserved=1;
	  break;
	case P_32FAT_1:
	  *reserved=32;
	  break;
  }
  switch(pos_fat_test)
  {
	case 0:
	  *fat_length=0;
	  break;
	case 1:
	  *fat_length=start_fat[0].offset-*reserved;
	  break;
	default:
	  *reserved=start_fat[0].offset;
	  *fat_length=start_fat[1].offset-*reserved;
	  break;
  }
  return part_type;
}

int search_fat16(t_param_disk disk_car,t_diskext partition,dword max_offset,int p_fat12,int p_fat16,int p_fat32,int debug,int dump_ind)
{
  int part_type;
  int dir_entries=0;
  int cluster_size;
  dword root_cluster=0;
  dword start_data,data_size;
  int reserved;
  int fat_length;
  if(debug)
	ecrit_rapport("search_fat16(partition,max_offset=%d,p_fat12=%d,p_fat16=%d,p_fat32=%d,debug=%d,dump_ind=%d)\n",max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);

  part_type=set_fat_offset(disk_car,&reserved, &fat_length, partition,max_offset,p_fat12,p_fat16,p_fat32,debug,dump_ind);
  partition->part_type=part_type;
  start_data=reserved+2*fat_length+(part_type==P_32FAT_1?0:((dir_entries-1)/16+1));
  data_size=partition->part_size-start_data;
  cluster_size=calcul_cluster_size(part_type,data_size,fat_length);
  switch(part_type)
  {
	case P_12FAT:
	case P_16FAT:
	  dir_entries=analyse_dir_entries(disk_car,partition,reserved+2*fat_length);
	  break;
	case P_32FAT_1:
	  root_cluster=get_root_cluster(disk_car,partition,cluster_size,data_size/cluster_size,start_data);
	  break;
	default:
	  ecrit_rapport("search_fat16 severe error\n");
	  return 1; /* Echec */
  }
  create_boot_sector(disk_car,partition, reserved, debug,dir_entries,root_cluster,cluster_size,fat_length);
  /*
  {
	int i;
	for(i=250;i<260;i++)
	  ecrit_rapport("\n%u=>%u",i,get_next_cluster(partition, part_type,reserved, i));
  }
  */
  return 0;
}

int recover_FAT_CIH(t_param_disk disk_car, t_diskext partition, const int debug, const int dump_ind)
{
  int res;
  dword fat_length_max;
//  dword start_rootdir,start_data,no_of_cluster,
//  int nb_entree_tbfat;
//  dword start_fat[MAX_TBFAT];
  int p_fat12,p_fat16,p_fat32;
  int cluster_size_min=2;
  if(partition->part_size>2*1024*1024*2+63)
  {
    p_fat32=1;
    p_fat16=0;
    p_fat12=0;
  }
  else
/* 1<<12 clusters * 8 secteurs/clusters= 32768 secteurs 
   fat_length=((1<<12+1)*1.5/512)+1=13; */
    if(partition->part_size>=(1+2*13+32768+63))
    {
      p_fat32=1;
      p_fat16=1;
      p_fat12=0;
    }
    else
    {
      p_fat32=0;
      p_fat16=1;
      p_fat12=1;
    }
  if(debug)
  {
    aff_part_rapport(1,disk_car,partition);
    ecrit_rapport("\nrecover_FAT_CIH p_fat12 %d, p_fat16 %d, p_fat32 %d\n", p_fat12,p_fat16,p_fat32);
  }
  /* Set fat_length_max */
  if(p_fat32)
  {	/* Cluster 1k */
    fat_length_max=partition->part_size/cluster_size_min*4/512;
  }
  else
  if(p_fat16)
  {
    while(partition->part_size/cluster_size_min>(1<<16))
      cluster_size_min*=2;
    fat_length_max=partition->part_size/cluster_size_min*2/512;
  }
  else
  {
    while(partition->part_size/cluster_size_min>(1<<12))
      cluster_size_min*=2;
    fat_length_max=partition->part_size/cluster_size_min*1.5/512;
  }
  if(debug)
    ecrit_rapport("fat_length_max %d\n", fat_length_max);
  res=search_fat16(disk_car,partition,64+fat_length_max,p_fat12,p_fat16,p_fat32,debug,dump_ind);
  if(debug)
    aff_part_rapport(1,disk_car,partition);
  return res;
}

int test_FAT(t_param_disk_cst disk_car,struct fat_boot_sector *fat_header, t_diskext_cst partition,const int debug, const int dump_ind)
{
  dword start_fat1,start_fat2,start_rootdir,start_data,no_of_cluster,fat_length,fat_length_calc,part_size,end_data;
  const char *buffer=(const char*)fat_header;
  if(fat_header->marker!=0xAA55)
  { aff_buffer(BUFFER_ADD,"test_FAT : " m_BAD_NMARK); return 1; }
  if(dump_ind)
    dump(fat_header,0x200);
  if(!((fat_header->ignored[0]==0xeb&&fat_header->ignored[2]==0x90)||fat_header->ignored[0]==0xe9))
  { aff_buffer(BUFFER_ADD,m_CHKFAT_BAD_JUMP); return 1; }
  if(sector_size(fat_header)!=512)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_BYTESECT); return 1; }
  switch(fat_header->cluster_size)
  { /* should depend of FAT12/16/32 */
    case 1:case 2:case 4:case 8:case 16:case 32:case 64:case 128: break;
    default: aff_buffer(BUFFER_ADD,m_CHKFAT_SECT_CLUSTER); return 1;
  }
  if(fat_header->secs_track!=disk_car->CHS.sector)
  {
    aff_buffer(BUFFER_ADD,m_CHKFAT_SECTTRACK);
    ecrit_rapport("sect/track %d %d\n",fat_header->secs_track,disk_car->CHS.sector);
    return 1;
  }
  if(fat_header->heads!=disk_car->CHS.head+1)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_SIDE); return 1; }
  if(fat_header->fats!=2)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_NBRFAT); return 1; }
  if(get_dir_entries(fat_header)%16)        /* Improved it */
  { aff_buffer(BUFFER_ADD,m_CHKFAT_ENTRY); return 1; }
  if(fat_header->media!=0xF8)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_MEDIA); return 1; }

  if(memcmp(buffer+FAT_NAME1,"FAT12",5)==0) /* 2 Mo max */
  {
    part_size=sectors(fat_header);
    if(fat_header->reserved!=1)
    { aff_buffer(BUFFER_ADD,m_CHKFAT_RESERV); return 1; }
    if(get_dir_entries(fat_header)==0)
    { aff_buffer(BUFFER_ADD,m_CHKFAT_ENTRY); return 1; }
    if(fat_header->fat_length>256)
    { aff_buffer(BUFFER_ADD,m_CHKFAT_SECTPFAT); return 1; }
    fat_length=fat_header->fat_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_rootdir=start_fat2+fat_length;
    start_data=start_rootdir+(get_dir_entries(fat_header)-1)/16+1;
    no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*1.5/512)+1;
  }
  else if(memcmp(buffer+FAT_NAME1,"FAT16",5)==0)
  {
    if(sectors(fat_header))
      part_size=sectors(fat_header);
    else
      part_size=fat_header->total_sect;
    fat_length=fat_header->fat_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_rootdir=start_fat2+fat_length;
    start_data=start_rootdir+(get_dir_entries(fat_header)-1)/16+1;
    no_of_cluster=(fat_header->total_sect-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*2/512)+1;
  }
  else if(memcmp(buffer+FAT_NAME2,"FAT32",5)==0)
  {
    part_size=fat_header->total_sect;
    if(sectors(fat_header)!=0)
    { aff_buffer(BUFFER_ADD,m_CHKFAT_SIZE); return 1; }
    fat_length=fat_header->fat_length?fat_header->fat_length:fat_header->fat32_length;
    start_fat1=fat_header->reserved;
    start_fat2=start_fat1+fat_length;
    start_data=start_fat2+fat_length;
    start_rootdir=start_data+fat_header->root_cluster*fat_header->cluster_size;
    no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
    fat_length_calc=((no_of_cluster+1)*4/512)+1;
  } else
  {
    ecrit_rapport("unknown part_type\n");
    return 1;
  }
  end_data=start_data+no_of_cluster*fat_header->cluster_size;
  if(partition->part_size)
  {
    if(part_size!=partition->part_size)
      ecrit_rapport("\ntest_FAT size boot_sector %d, partition %d\n",part_size,partition->part_size);
    if(part_size>partition->part_size)
    { aff_buffer(BUFFER_ADD,m_CHKFAT_SIZE); return 1; }
  }
  if(debug)
  {
    ecrit_rapport("\n");
    aff_part_rapport(2,disk_car,partition);
    ecrit_rapport("FAT1 : %u-%u\n",start_fat1,start_fat1+fat_length-1);
    ecrit_rapport("FAT2 : %u-%u\n",start_fat2,start_fat2+fat_length-1);
    ecrit_rapport("start_rootdir : %u\n",start_rootdir);
    ecrit_rapport("Data : %u-%u\n",start_data,end_data);
    ecrit_rapport("sectors : %u\n",part_size);
    ecrit_rapport("cluster_size : %u\n",fat_header->cluster_size);
    ecrit_rapport("no_of_cluster : %u (2 - %u)\n", no_of_cluster,no_of_cluster+1);
    ecrit_rapport("fat_length %u calculated %u\n",fat_length,fat_length_calc);
  }
  if(fat_length<fat_length_calc)
  { aff_buffer(BUFFER_ADD,m_CHKFAT_SECTPFAT); return 1; }
//  if(fat_header->hidden!=partition->lba)
//  {	/* mkdosfs has this little bug */
//    aff_buffer(BUFFER_ADD,m_CHKFAT_BAD_HIDDEN);
//    ecrit_rapport("Incorrect number of hidden sectors: fat_header->hidden=%d LBA=%d\n",fat_header->hidden,partition->lba);
//    return 1;
//  }
  return comp_FAT(disk_car,partition,fat_length,fat_header->reserved);
}

int comp_FAT(t_param_disk_cst disk_car,t_diskext_cst partition,const dword taille,const dword sect_res)
{
  uint reste=taille;
  uint read_size;
  uchar buffer[NBR_SECT*0x200],buffer2[NBR_SECT*0x200];
  loff_t hd_offset, hd_offset2;
  hd_offset=partition->lba+sect_res;
  hd_offset2=hd_offset+taille;
  if(reste>1000) reste=1000;	/* Quick check ! */
  while(reste)
  {
    read_size=reste>NBR_SECT?NBR_SECT:reste;
    reste-=read_size;
    if(hd_read_LBA(disk_car,read_size, &buffer, hd_offset))
    { ecrit_rapport(m_CHKFAT_RERR);
      return 1;}
    if(hd_read_LBA(disk_car,read_size, &buffer2, hd_offset2))
    { ecrit_rapport(m_CHKFAT_RERR); 
      return 1;}
    if(memcmp(buffer,buffer2,0x200*read_size)!=0)
    { ecrit_rapport("FAT differs, FAT sector=%d-%d/%d\n",taille-reste,taille-reste+read_size,taille); 
      return 1;
    }
//    if(hd_offset==partition->lba+sect_res)
//      dump(buffer,0x200);
    hd_offset+=read_size;
    hd_offset2+=read_size;
  }
  return 0;
}

uint sector_size(const struct fat_boot_sector *fat_header)
{ return (fat_header->sector_size[1]<<8)+fat_header->sector_size[0]; }

uint get_dir_entries(const struct fat_boot_sector *fat_header)
{ return (fat_header->dir_entries[1]<<8)+fat_header->dir_entries[0]; }

uint sectors(const struct fat_boot_sector *fat_header)
{ return (fat_header->sectors[1]<<8)+fat_header->sectors[0]; }

int recover_FAT12(t_param_disk disk_car,struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT12",5)==0) /* 2 Mo max */
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT12 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
//      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
//	return 1;
      partition->part_size=sectors(fat_header);
      partition->part_type=P_12FAT;
      set_part_name(partition,buffer+FAT1X_PART_NAME,11);
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_FAT16(t_param_disk disk_car,struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT16",5)==0)
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT16 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
	return 1;
      if(sectors(fat_header)!=0)
      {
	partition->part_size=sectors(fat_header);
	partition->part_type=P_16FAT;
      } else {
	partition->part_size=fat_header->total_sect;
	partition->part_type=P_16FATBD;
      }
      set_part_name(partition,buffer+FAT1X_PART_NAME,11);
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_FAT32(t_param_disk disk_car,struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME2,"FAT32",5)==0)
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nFAT32 at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
//      if(test_FAT(disk_car, fat_header, partition, debug, dump_ind))
//	return 1;
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      partition->part_type=P_32FAT_1;
      set_FAT32_part_name(disk_car,partition,fat_header);
      return 0;
    }
  }     /* fin marqueur de fin =:-) */
  return 1;
}

int set_FAT32_part_name(t_param_disk_cst disk_car, t_diskext partition, struct fat_boot_sector*fat_header)
{
  int i;
  unsigned char buffer2[0x200*fat_header->cluster_size];
  if(hd_read_LBA(disk_car,fat_header->cluster_size, &buffer2, partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat32_length+(fat_header->root_cluster-2)*fat_header->cluster_size))
  {
    ecrit_rapport(m_ROOT_CLUSTER_RERR); return 1;
  }
  for(i=0;(i<16*fat_header->cluster_size);i++)
  {
    if(buffer2[i*0x20+0xB]==0x28) /* Test attribut volume name */
    {
      set_part_name(partition,&buffer2[i*0x20],11);
      return 0;
    }
  }
  return 0;
}

int recover_HPFS(t_param_disk disk_car,struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+OS2_NAME,"IBM",3)==0)
    {   /* D'apres une analyse de OS2 sur systeme FAT...
	   FAT_NAME1=FAT
	 */
      if(debug||dump_ind)
      {
	ecrit_rapport("\nHPFS maybe at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind)
	dump(buffer,0x200);
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      partition->part_type=P_HPFS;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int recover_OS2MB(t_param_disk disk_car,struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
    if(memcmp(buffer+FAT_NAME1,"FAT",3)==0)
    {   /* Attention a detecter apres OS2 */
      if(debug||dump_ind)
      {
	ecrit_rapport("\nMarker (0xAA55) at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind)
	dump(buffer,0x200);
      partition->part_size=(disk_car->CHS.head+1) * disk_car->CHS.sector; /* 1 cylinder */
      partition->part_type=P_OS2MB;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}
