/*

    File: fnctdsk.c

    Copyright (C) 1998-2001  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


#include <stdlib.h>
#ifdef DJGPP
#include <curses.h>
#else
#include <ncurses.h>
#include <sys/types.h>
#endif
#include <stdio.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "fnctdsk.h"
#include "lang.h"
#include "testdisk.h"
#include "hdaccess.h"
#include "interface.h"
#include "analyse.h"
extern char* nom_os[256];

/* BOOT */
/*
dword NT_sect(t_sector_cst B)
{ return *((const dword*)&(B[0x28]));
}

dword NT_MFT_cluster(t_sector_cst B)
{ return *((const dword*)&(B[0x30]));
}

byte NT_clustersize(t_sector_cst B)
{ return B[0x0D];
}
*/

int bootable(t_entree_cst E)
{
  return E[0];
}

dword sect_rel(t_entree_cst E)
{ return *((const dword*)&(E[8]));
}

dword nbr_sect(t_entree_cst E)
{ return *((const dword*)&(E[12]));
}

dword CHS2LBA(t_param_disk_cst disk_car,const t_CHS*CHS)
{ return ((dword)CHS->cylinder*(disk_car->CHS.head+1)+CHS->head)*disk_car->CHS.sector+CHS->sector-1;
}

dword C_H_S2LBA(t_param_disk_cst disk_car,const word C, const byte H, const byte S)
{ return ((dword)C*(disk_car->CHS.head+1)+H)*disk_car->CHS.sector+S-1;
}

dword get_LBA(t_param_disk_cst disk_car,t_disk_cst pos)
{
  return C_H_S2LBA(disk_car,pos->cylinder, pos->head, pos->sector);
}

t_disk set_pos(t_param_disk_cst disk_car,t_disk pos, dword sect)
{
  pos->sector=(sect%disk_car->CHS.sector)+1;
  sect/=disk_car->CHS.sector;
  pos->head=sect%(disk_car->CHS.head+1);
  pos->cylinder=sect/(disk_car->CHS.head+1);
  return pos;
}

dword LBA2sector(t_param_disk_cst disk_car, dword sect)
{ return (sect%disk_car->CHS.sector)+1; }

dword LBA2head(t_param_disk_cst disk_car, dword sect)
{ return (sect/disk_car->CHS.sector)%(disk_car->CHS.head+1); }

dword LBA2cylinder(t_param_disk_cst disk_car, dword sect)
{ return (sect/disk_car->CHS.sector)/(disk_car->CHS.head+1); }

void LBA2CHS(t_param_disk_cst disk_car,const dword lba, t_CHS*CHS)
{
  CHS->cylinder=LBA2cylinder(disk_car,lba);
  CHS->head=LBA2head(disk_car,lba);
  CHS->sector=LBA2sector(disk_car,lba);
}

void set_pos2(t_param_disk_cst disk_car, t_CHS * CHS, const dword sect)
{
  dword pos=sect;
  CHS->sector=(pos%disk_car->CHS.sector)+1;
  pos/=disk_car->CHS.sector;
  CHS->head=pos%(disk_car->CHS.head+1);
  CHS->cylinder=pos/(disk_car->CHS.head+1);
}

dword get_sect_s(t_param_disk_cst disk_car,t_entree_cst E)
{ return C_H_S2LBA(disk_car,s_cyl(E),s_head(E),s_sect(E));
}

dword get_sect_e(t_param_disk_cst disk_car,t_entree_cst E)
{ return C_H_S2LBA(disk_car,e_cyl(E),e_head(E),e_sect(E));
}

void dup_t_CHS(t_CHS * CHS_dest, const t_CHS * CHS_source)
{
  CHS_dest->cylinder=CHS_source->cylinder;
  CHS_dest->head=CHS_source->head;
  CHS_dest->sector=CHS_source->sector;
}

int compare2uint(uint a, uint b)
{
  if(a<b)
    return -1;
  else
    if(a==b)
      return 0;
    else
      return 1;
}

int compare_CHS(const t_CHS *CHS1, const t_CHS *CHS2)
{
  int res;
  res=compare2uint(CHS1->cylinder,CHS2->cylinder);
  if(res!=0)
    return res;
  res=compare2uint(CHS1->head,CHS2->head);
  if(res!=0)
    return res;
  return compare2uint(CHS1->sector,CHS2->sector);
}

void dup_t_param_disk(t_param_disk dest, const t_param_disk src)
{
  dup_t_CHS(&dest->CHS,&src->CHS);
  dest->lba=src->lba;
  dest->disk=src->disk;
#ifdef DJGPP
  dest->mode=src->mode;
  dest->nbr_err=src->nbr_err;
  dup_t_CHS(&dest->CHSR,&src->CHSR);
#else
  dest->name=src->name;
#endif
}

void dup_t_diskext(t_diskext dest, t_diskext_cst src)
{
  dest->lba=src->lba;
  dest->disk=src->disk;
  dest->part_type=src->part_type;
  dest->part_size=src->part_size;
  dest->status=src->status;
  dest->order=src->order;
  strncpy(dest->info,src->info,sizeof(dest->info));
  strncpy(dest->name,src->name,sizeof(dest->name));
}

void partition2entree(t_param_disk_cst disk_car,dword pos,t_diskext_cst partition, t_entree *entree)
{
  t_CHS start,end;
  set_pos2(disk_car,&start,partition->lba);
  set_pos2(disk_car,&end,partition->lba+partition->part_size-1);
  if(partition->status==STATUS_PRIM_BOOT)
    (*entree)[0]=0x80;
  else
    (*entree)[0]=0;             /* Non bootable */
  (*entree)[4]=partition->part_type;
  *((dword*)&((*entree)[8]))=partition->lba-pos;
  if(start.cylinder>1023)
  { /* Partition Magic 5 uses CHS=(1023,0,1) if extended or last logical *
     * Linux fdisk and TestDisk use CHS=(1023,254,63)                    */
    (*entree)[1]=254;
    (*entree)[2]=63|((1023>>8)<<6);
    (*entree)[3]=(uchar)1023;
  }
  else
  {
    (*entree)[1]=start.head;
    (*entree)[2]=start.sector|((start.cylinder>>8)<<6);
    (*entree)[3]=start.cylinder;
  }
  if(end.cylinder>1023)
  {
    (*entree)[5]=254;
    (*entree)[6]=63|((1023>>8)<<6);
    (*entree)[7]=(uchar)1023;
  }
  else
  {
    (*entree)[5]=end.head;
    (*entree)[6]=end.sector|((end.cylinder>>8)<<6);
    (*entree)[7]=end.cylinder;
  }
  *((dword*)&((*entree)[12]))=partition->part_size;
}

int entree2partition(t_param_disk_cst disk_car,dword pos,t_diskext partition, t_entree_cst entree, const int status,const int order,const int debug)
{
  t_CHS start,end;
  partition->disk=disk_car->disk;
  partition->part_type=os(entree);
  partition->lba=pos+sect_rel(entree);
  partition->order=order;
  partition->part_size=nbr_sect(entree);
  partition->info[0]='\0';
  partition->name[0]='\0';

  start.cylinder=s_cyl(entree);
  start.head=s_head(entree);
  start.sector=s_sect(entree);
  end.cylinder=e_cyl(entree);
  end.head=e_head(entree);
  end.sector=e_sect(entree);

  switch(status)
  {
    case STATUS_PRIM:
      if(partition->part_type==P_EXTENDX || partition->part_type==P_EXTENDED)
	partition->status=STATUS_EXT;
      else
	if(bootable(entree))
	  partition->status=STATUS_PRIM_BOOT;
	else
	  partition->status=status;
      break;
    default:
      partition->status=status;
      break;
  }
  if(start.cylinder<1023)
  {
    if(partition->lba!=get_sect_s(disk_car,entree))
    {
      ecrit_rapport("BAD_RS LBA=%d %d\n",partition->lba,get_sect_s(disk_car,entree));
      partition->errcode=BAD_RS;
      return 1;
    }
  }

  if(end.cylinder<1023)
  {
    dword end_sect;
    end_sect=partition->lba+partition->part_size-1;
    if(end_sect!=get_sect_e(disk_car,entree))
    {
      partition->errcode=BAD_SCOUNT;
      return 1;
    }
  }
  /* Check CHS */
  if((start.sector==0)||(start.sector>disk_car->CHS.sector))
  {
    partition->errcode=BAD_SS;
    return 1;
  }
  if((end.sector==0)||(end.sector>disk_car->CHS.sector))
  {
    partition->errcode=BAD_ES;
    return 1;
  }
  if(start.head>disk_car->CHS.head)
  {
    partition->errcode=BAD_SH;
    return 1;
  }
  if(start.cylinder>disk_car->CHS.cylinder)
  {
    partition->errcode=BAD_SC;
    return 1;
  }
  if(end.head>disk_car->CHS.head)
  {
    partition->errcode=BAD_EH;
    return 1;
  }
  if(end.cylinder>disk_car->CHS.cylinder)
  {
    partition->errcode=BAD_EC;
    return 1;
  }
  /* Check start < end */
  if(start.cylinder>end.cylinder)
  {
    partition->errcode=BAD_EBS;
    return 1;
  }
  if(start.cylinder==end.cylinder)
  {
    if(start.head>end.head)
    {
      partition->errcode=BAD_EBS;
      return 1;
    }
    if(start.head==end.head)
      if(start.sector>end.sector)
      {
	partition->errcode=BAD_EBS;
	return 1;
      }
  }
  /* Check partition and load partition name */
  check_part(disk_car,debug,partition);
  return 0;
}

const char* errmsg_entree2partition(const unsigned int errcode)
{
  switch(errcode)
  {
    case BAD_SS: return m_BAD_S_SECT;
    case BAD_ES: return m_BAD_E_SECT;
    case BAD_SH: return m_BAD_S_HEAD;
    case BAD_EH: return m_BAD_E_HEAD;
    case BAD_EBS: return m_END_BFR_START;
    case BAD_RS: return m_BAD_RS;
    case BAD_SC: return m_BAD_S_CYL;
    case BAD_EC: return m_BAD_E_CYL;
    case BAD_SCOUNT: return m_BAD_SCOUNT;
  }
  return "";
}

int read_MBR(t_param_disk disk_car,void *buffer)
{
  if(hd_read_LBA(disk_car,1, buffer, 0))
  {
    doprintf(m_PART_RD_ERR);
    return 1;
  }
  return 0;
}

int write_MBR(t_param_disk disk_car,void *buffer)
{
  if(hd_write_LBA(disk_car,1, buffer, 0))
  {
    doprintf(m_PART_WR_ERR);
    return 1;
  }
  return 0;
}

t_list_part insert_new_partition(t_list_part list_part, t_list_part new)
{
  t_list_part prev,next;
  prev=NULL;
  for(next=list_part;;next=next->next)
  { // prev new next
    if(!next||(new->part->lba<next->part->lba)||((new->part->lba==next->part->lba)&&(new->part->part_size<=next->part->part_size)))
    {
      if(next&&(next->part->lba==new->part->lba)&&(next->part->part_size==new->part->part_size)&&(next->part->part_type==new->part->part_type))
	return list_part;
      else
      { /* new current */
	new->next=next;
	new->prev=prev;
	if(next)
	  next->prev=new;
	if(prev)
	{
	  prev->next=new;
	  return list_part;
	}
	return new;
      }
    }
    prev=next;
  }
}

int check_list_part(t_list_part list_part)
{
  t_list_part prev=NULL,parts;
  if(list_part && list_part->prev)
  {
    ecrit_rapport("\ncheck_list_part error\n");
    exit(1);
  }
  for(parts=list_part;parts;parts=parts->next)
  {
//	ecrit_rapport("%p %p %p\n",parts->prev, parts, parts->next);
	if(prev!=parts->prev)
	{
	  ecrit_rapport("\ncheck_list_part error\n");
	  exit(1);
	}
	prev=parts;
  }
  if(prev && prev->next)
  {
    ecrit_rapport("\ncheck_list_part error\n");
    exit(1);
  }
  return 0;
}

t_list_part sort_list_part(t_list_part list_part)
{
  t_list_part new_list_part=NULL,parts,next;
  for(parts=list_part;parts;parts=next)
  {
    next=parts->next;
    new_list_part=insert_new_partition(new_list_part,parts);
  }
  return new_list_part;
}
